
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"

extern void	Error(),RangeList(),ReadSliceImage();

static int	R_D(str,N)
char		*str;
int		N;
{
	int	D=(*str=='-')?(int)(atof(str+1)*(double)N+0.5):atoi(str);

	if (D<=0 || D>N) Error("bad correlation distance.");

	return D;
}

#define LEN	1024
#define BINARY	unsigned char
#define CELL	(size_t)((Cell)(~0))
#define COUNT	unsigned long

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,form[LEN],path[LEN];
	int	Nx,Ny,Nz,Dx,Dy,Dz,BPS,
		Mx,X,dx,My,Y,dy,Mz,Z,y,z,x,z0,y1,y2,x1,x2,dz,l;
	BINARY	binary[CELL+1],***B;
	Cell	**cell,c;
	COUNT	***C;
	double	p,q,r,s,t;

	if (argc!=6 && argc!=9)
	Error("usage : si_2cc orgDir nameFile rangeList {Dx Dy Dz} BPS newDir");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	(void)printf("%d\t%d\t%d\n",Nx,Ny,Nz);

	RangeList(argv[3],CELL,binary);

	if (argc==6) {
	    Dx=Nx; Dy=Ny; Dz=Nz;
	}
	else {
	    Dx=R_D(argv[4],Nx);
	    Dy=R_D(argv[5],Ny);
	    Dz=R_D(argv[6],Nz);
	}
	(void)printf("%d\t%d\t%d\n",Dx,Dy,Dz);

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad BPS.");

	Mx=Dx*2-1; X=1-Dx; dx=(Nx>Mx)?Nx:Mx;
	My=Dy*2-1; Y=1-Dy; dy=(Ny>My)?Ny:My;
	Mz=Dz*2-1; Z=1-Dz;

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

	if ((cell=ALLOC(Cell *,(size_t)dy))==NULL ||
	    (*cell=ALLOC(Cell,(size_t)dy*(size_t)dx))==NULL ||
	    (B=ALLOC(BINARY **,(size_t)Nz))==NULL ||
	    (*B=ALLOC(BINARY *,(size_t)Dz*(size_t)Ny))==NULL ||
	    (**B=ALLOC(BINARY,(size_t)Dz*(size_t)Ny*(size_t)Nx))==NULL ||
	    (C=ALLOC(COUNT **,(size_t)Dz))==NULL ||
	    (*C=ALLOC(COUNT *,(size_t)Dz*(size_t)My))==NULL ||
	    (**C=ALLOC(COUNT,(size_t)Dz*(size_t)My*(size_t)Mx))==NULL)
	    Error("memory allocation error.");

	for (y=1; y<dy; y++) cell[y]=cell[y-1]+dx;

	for (y=1; y<Ny; y++) B[0][y]=B[0][y-1]+Nx;
	for (z=1; z<Dz; z++) {
	    B[z]=B[z-1]+Ny; for (y=0; y<Ny; y++) B[z][y]=B[z][y-1]+Nx;
	}
	for (; z<Nz; z++) B[z]=B[z%Dz];

	C-=Z; C[Z]-=Y; C[Z][Y]-=X;
	for (y=Y+1; y<Dy; y++) {
	    C[Z][y]=C[Z][y-1]+Mx; for (x=X; x<Dx; x++) C[Z][y][x]=0;
	}
	for (z=Z+1; z<=0; z++) {
	    C[z]=C[z-1]+My;
	    for (y=Y; y<Dy; y++) {
		C[z][y]=C[z][y-1]+Mx; for (x=X; x<Dx; x++) C[z][y][x]=0;
	    }
	}
	for (z=0; z<Nz; z++) {
	    ReadSliceImage(paths[z],Nx,Ny,cell);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) B[z][y][x]=binary[cell[y][x]];

	    z0=(z+Z>=0)?Z:-z;
	    for (y=0; y<Ny; y++) {
		y1=(y+Y>=0)?Y:-y; y2=(y+Dy<=Ny)?Dy:Ny-y;
		for (x=0; x<Nx; x++) if (B[z][y][x]) {
		    x1=(x+X>=0)?X:-x; x2=(x+Dx<=Nx)?Dx:Nx-x;
		    for (dz=z0; dz<=0; dz++)
		    for (dy=y1; dy<y2; dy++)
		    for (dx=x1; dx<x2; dx++) C[dz][dy][dx]+=B[z+dz][y+dy][x+dx];
		}
	    }
	}
	if (***C==0) Error("no object pixel.");

	(void)printf("%lf\n",p=(double)***C/(double)Nx/(double)Ny/(double)Nz);

	l=1; for (z=10; z<Mz; z*=10) ++l;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[argc-1],l);

	q=(double)(c=(Cell)((1<<BPS)-1))/p;

#define PV(Cxyz)	\
	((t=(double)Cxyz*s/(double)(Nx-abs(x)))<=(double)c)?(Cell)(t+0.5):c

	for (z=Z; z<=0; z++) {
	    r=q/(double)(Nz+z);
	    for (y=Y; y<Dy; y++) {
		s=r/(double)(Ny-abs(y));
		for (x=X; x<Dx; x++) cell[y-Y][x-X]=PV(C[z][y][x]);
	    }
	    (void)sprintf(path,form,z-Z);
	    StoreImageFile(path,Mx,My,BPS,cell,NULL);
	}
	for (; z<Dz; z++) {
	    r=q/(double)(Nz-z);
	    for (y=Y; y<Dy; y++) {
		s=r/(double)(Ny-abs(y));
		for (x=X; x<Dx; x++) cell[y-Y][x-X]=PV(C[-z][-y][-x]);
	    }
	    (void)sprintf(path,form,z-Z);
	    StoreImageFile(path,Mx,My,BPS,cell,NULL);
	}
	return 0;
}
