
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"
#include "FFT.h"

extern void	Error(),RangeList(),ReadSliceImage();

static int	R_D(str,N)
char		*str;
int		N;
{
	int	D=(*str=='-')?(int)(atof(str+1)*(double)N+0.5):atoi(str);

	if (D<=0 || D>N) Error("bad correlation distance.");

	return D;
}

#define CELL	(size_t)((Cell)(~0))
#define LEN	1024

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,binary[CELL+1],form[LEN],path[LEN];
	int	Nx,Ny,Nz,Dx,Dy,Dz,BPS,X,Mx,Y,My,Lx,Ly,Lz,Ix,y,z,x,Z,l;
	Cell	**cell,c;
	Complex	*Ex,*Ey,*Ez,*W,***F;
	double	***P,Lxyz,p,q,r,s,t;

	if (argc!=5 && argc!=6 && argc!=8 && argc!=9)
    Error("usage : si_acc orgDir nameFile {rangeList} {Dx Dy Dz} BPS newDir");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	(void)printf("%d\t%d\t%d\n",Nx,Ny,Nz);

	if (argc==6 || argc==9) RangeList(argv[3],CELL,binary);

	if (argc==5 || argc==6) {
	    Dx=Nx; Dy=Ny; Dz=Nz;
	}
	else {
	    Dx=R_D(argv[argc-5],Nx);
	    Dy=R_D(argv[argc-4],Ny);
	    Dz=R_D(argv[argc-3],Nz);
	}
	(void)printf("%d\t%d\t%d\n",Dx,Dy,Dz);

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad BPS.");

	X=(Nx>(Mx=Dx*2-1))?Nx:Mx; Y=(Ny>(My=Dy*2-1))?Ny:My;

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

	if ((cell=ALLOC(Cell *,(size_t)Y))==NULL ||
	    (*cell=ALLOC(Cell,(size_t)Y*(size_t)X))==NULL ||
	    (Lx=SetUpFFT(Nx+Dx,&Ex))==0 ||
	    (Ly=SetUpFFT(Ny+Dy,&Ey))==0 ||
	    (Lz=SetUpFFT(Nz+Dz,&Ez))==0 ||
    (W=ALLOC(Complex,(size_t)((Lx>Ly)?(Lx>Lz)?Lx:Lz:(Ly>Lz)?Ly:Lz)))==NULL ||
	    (F=ALLOC(Complex **,(size_t)Nz))==NULL ||
	    (*F=ALLOC(Complex *,(size_t)Nz*(size_t)Ly))==NULL ||
	(**F=ALLOC(Complex,(size_t)Nz*(size_t)Ly*(size_t)(Ix=Lx/2+1)))==NULL)
	    Error("memory allocation error.");

	for (y=1; y<Y; y++) cell[y]=cell[y-1]+X;

	for (y=1; y<Ly; y++) F[0][y]=F[0][y-1]+Ix;
	for (z=1; z<Nz; z++) {
	    F[z]=F[z-1]+Ly; for (y=0; y<Ly; y++) F[z][y]=F[z][y-1]+Ix;
	}
	for (z=0; z<Nz; z++) {
	    ReadSliceImage(paths[z],Nx,Ny,cell);

	    if (argc==6 || argc==9)
		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) cell[y][x]=(Cell)binary[cell[y][x]];

	    for (y=0; y<Ny; y++) {
		for (x=0; x<Nx; x++) {
		    W[x].r=(double)cell[y][x]; W[x].i=0.0;
		}
		for (; x<Lx; x++) W[x].r=W[x].i=0.0;

		FFT(Lx,Ex,W,-1);

		for (x=0; x<Ix; x++) F[z][y][x]=W[x];
	    }
	    for (x=0; x<Ix; x++) {
		for (y=0; y<Ny; y++) W[y]=F[z][y][x];
		for (; y<Ly; y++) W[y].r=W[y].i=0.0;

		FFT(Ly,Ey,W,-1);

		for (y=0; y<Ly; y++) F[z][y][x]=W[y];
	    }
	}
	for (y=0; y<Ly; y++)
	for (x=0; x<Ix; x++) {
	    for (z=0; z<Nz; z++) W[z]=F[z][y][x];
	    for (; z<Lz; z++) W[z].r=W[z].i=0.0;

	    FFT(Lz,Ez,W,-1);

	    for (z=0; z<Lz; z++) {
		W[z].r=W[z].r*W[z].r+W[z].i*W[z].i; W[z].i=0.0;
	    }
	    FFT(Lz,Ez,W,1);

	    for (z=0; z<Dz; z++) F[z][y][x]=W[z];
	}
	P=(double ***)F; Lxyz=(double)Lx*(double)Ly*(double)Lz;

	for (z=0; z<Dz; z++) {
	    for (x=0; x<Ix; x++) {
		for (y=0; y<Ly; y++) W[y]=F[z][y][x];

		FFT(Ly,Ey,W,1);

		for (y=0; y<Ly; y++) F[z][y][x]=W[y];
	    }
	    for (y=0; y<Ly; y++) {
		for (x=0; x<Ix; x++) W[x]=F[z][y][x];
		for (; x<Lx; x++) {
		    W[x].r=F[z][y][Lx-x].r; W[x].i=(-F[z][y][Lx-x].i);
		}
		FFT(Lx,Ex,W,1);

		for (x=0; x<Lx; x++) P[z][y][x]=W[x].r/Lxyz;
	    }
	}
	if (***P==0.0) Error("no object pixel.");

	(void)printf("%lf\n",p=(double)***P/(double)Nx/(double)Ny/(double)Nz);

	X=1-Dx; Y=1-Dy; Z=1-Dz; q=(double)(c=(Cell)((1<<BPS)-1))/p;

	l=1; for (z=Dz*2-2; z>=10; z/=10) ++l;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[argc-1],l);

#define PV(Pxyz,nx)	((t=Pxyz*s/(double)(nx))<=(double)c)?(Cell)(t+0.5):c

	for (z=Z; z<0; z++) {
	    r=q/(double)(Nz+z);
	    for (y=Y; y<=0; y++) {
		s=r/(double)(Ny+y);
		for (x=X; x<=0; x++) cell[y-Y][x-X]=PV(P[-z][-y][-x],Nx+x);
		for (; x<Dx; x++) cell[y-Y][x-X]=PV(P[-z][-y][Lx-x],Nx-x);
	    }
	    for (; y<Dy; y++) {
		s=r/(double)(Ny-y);
		for (x=X; x<=0; x++) cell[y-Y][x-X]=PV(P[-z][Ly-y][-x],Nx+x);
		for (; x<Dx; x++) cell[y-Y][x-X]=PV(P[-z][Ly-y][Lx-x],Nx-x);
	    }
	    (void)sprintf(path,form,z-Z);
	    StoreImageFile(path,Mx,My,BPS,cell,NULL);
	}
	for (; z<Dz; z++) {
	    r=q/(double)(Nz-z);
	    for (y=Y; y<0; y++) {
		s=r/(double)(Ny+y);
		for (x=X; x<0; x++) cell[y-Y][x-X]=PV(P[z][Ly+y][Lx+x],Nx+x);
		for (; x<Dx; x++) cell[y-Y][x-X]=PV(P[z][Ly+y][x],Nx-x);
	    }
	    for (; y<Dy; y++) {
		s=r/(double)(Ny-y);
		for (x=X; x<0; x++) cell[y-Y][x-X]=PV(P[z][y][Lx+x],Nx+x);
		for (; x<Dx; x++) cell[y-Y][x-X]=PV(P[z][y][x],Nx-x);
	    }
	    (void)sprintf(path,form,z-Z);
	    StoreImageFile(path,Mx,My,BPS,cell,NULL);
	}
	return 0;
}
