
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"
#include "FFT.h"

extern void	Error(),RangeList(),ReadSliceImage();

static int	R_D(str,N)
char		*str;
int		N;
{
	int	D=(*str=='-')?(int)(atof(str+1)*(double)N+0.5):atoi(str);

	if (D<=0 || D>N) Error("bad correlation distance.");

	return D;
}

#define CELL	(size_t)((Cell)(~0))
#define LEN	1024

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,binary[CELL+1],tmp[LEN],tif[LEN],path[LEN];
	int	Nx,Ny,Nz,Dx,Dy,Dz,BPS,X,Mx,Y,My,Lx,Ly,Lz,Z,Ix,y,l,z,x;
	Cell	**cell,c;
	Complex	*Ex,*Ey,*Ez,*W,**F;
	FILE	*file;
	long	offset;
	double	**P,Lxyz,p,q,r,s,t;

	if (argc!=5 && argc!=6 && argc!=8 && argc!=9)
    Error("usage : si_acc orgDir nameFile {rangeList} {Dx Dy Dz} BPS newDir");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	(void)printf("%d\t%d\t%d\n",Nx,Ny,Nz);

	if (argc==6 || argc==9) RangeList(argv[3],CELL,binary);

	if (argc==5 || argc==6) {
	    Dx=Nx; Dy=Ny; Dz=Nz;
	}
	else {
	    Dx=R_D(argv[argc-5],Nx);
	    Dy=R_D(argv[argc-4],Ny);
	    Dz=R_D(argv[argc-3],Nz);
	}
	(void)printf("%d\t%d\t%d\n",Dx,Dy,Dz);

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad BPS.");

	X=(Nx>(Mx=Dx*2-1))?Nx:Mx; Y=(Ny>(My=Dy*2-1))?Ny:My;

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

	if ((cell=ALLOC(Cell *,(size_t)Y))==NULL ||
	    (*cell=ALLOC(Cell,(size_t)Y*(size_t)X))==NULL ||
	    (Lx=SetUpFFT(Nx+Dx,&Ex))==0 ||
	    (Ly=SetUpFFT(Ny+Dy,&Ey))==0 ||
	    (Lz=SetUpFFT(Nz+Dz,&Ez))==0 ||
    (W=ALLOC(Complex,(size_t)((Lx>Ly)?(Lx>Lz)?Lx:Lz:(Ly>Lz)?Ly:Lz)))==NULL ||
	    (F=ALLOC(Complex *,(size_t)(Z=(Ly>Nz)?Ly:Nz)))==NULL ||
	    (*F=ALLOC(Complex,(size_t)Z*(size_t)(Ix=Lx/2+1)))==NULL)
	    Error("memory allocation error.");

	for (y=1; y<Y; y++) cell[y]=cell[y-1]+X;

	for (y=1; y<Z; y++) F[y]=F[y-1]+Ix;

	l=1; for (z=Dz*2-2; z>=10; z/=10) ++l;
	(void)sprintf(tmp,"%s/%%0%dd.tmp",argv[argc-1],l);
	(void)sprintf(tif,"%s/%%0%dd.tif",argv[argc-1],l);

#define RW(ptr,noe)	(void *)ptr,sizeof(Complex)*noe,(size_t)1

	for (z=0; z<Nz; z++) {
	    ReadSliceImage(paths[z],Nx,Ny,cell);

	    if (argc==6 || argc==9)
		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) cell[y][x]=(Cell)binary[cell[y][x]];

	    for (y=0; y<Ny; y++) {
		for (x=0; x<Nx; x++) {
		    W[x].r=(double)cell[y][x]; W[x].i=0.0;
		}
		for (; x<Lx; x++) W[x].r=W[x].i=0.0;

		FFT(Lx,Ex,W,-1);

		for (x=0; x<Ix; x++) F[y][x]=W[x];
	    }
	    for (x=0; x<Ix; x++) {
		for (y=0; y<Ny; y++) W[y]=F[y][x];
		for (; y<Ly; y++) W[y].r=W[y].i=0.0;

		FFT(Ly,Ey,W,-1);

		for (y=0; y<Ly; y++) F[y][x]=W[y];
	    }
	    (void)sprintf(path,tmp,z);
	    if ((file=fopen(path,"wb"))==NULL ||
		fwrite(RW(*F,(size_t)Ly*(size_t)Ix),file)!=1 ||
		fclose(file))
		Error("temporary file write error.");
	}
	for (y=0; y<Ly; y++) {
	    offset=(long)sizeof(Complex)*(long)y*(long)Ix;

	    for (z=0; z<Nz; z++) {
		(void)sprintf(path,tmp,z);
		if ((file=fopen(path,"rb"))==NULL ||
		    fseek(file,offset,SEEK_SET) ||
		    fread(RW(F[z],(size_t)Ix),file)!=1 ||
		    fclose(file))
		    Error("temporary file read error.");
	    }
	    for (x=0; x<Ix; x++) {
		for (z=0; z<Nz; z++) W[z]=F[z][x];
		for (; z<Lz; z++) W[z].r=W[z].i=0.0;

		FFT(Lz,Ez,W,-1);

		for (z=0; z<Lz; z++) {
		    W[z].r=W[z].r*W[z].r+W[z].i*W[z].i; W[z].i=0.0;
		}
		FFT(Lz,Ez,W,1);

		for (z=0; z<Dz; z++) F[z][x]=W[z];
	    }
	    for (z=0; z<Dz; z++) {
		(void)sprintf(path,tmp,z);
		if ((file=fopen(path,"rb+"))==NULL ||
		    fseek(file,offset,SEEK_SET) ||
		    fwrite(RW(F[z],(size_t)Ix),file)!=1 ||
		    fclose(file))
		    Error("temporary file re-write error.");
	    }
	}
	for (z=Dz; z<Nz; z++) {
	    (void)sprintf(path,tmp,z);
	    (void)remove(path);
	}
	P=(double **)F; Lxyz=(double)Lx*(double)Ly*(double)Lz;
	X=1-Dx; Y=1-Dy; Z=1-Dz; c=(Cell)((1<<BPS)-1);

#define PV(Pxy,nx)	((t=Pxy*s/(double)(nx))<=(double)c)?(Cell)(t+0.5):c

	for (z=0; z<Dz; z++) {
	    (void)sprintf(path,tmp,z);
	    if ((file=fopen(path,"rb"))==NULL ||
		fread(RW(*F,(size_t)Ly*(size_t)Ix),file)!=1 ||
		fclose(file))
		Error("temporary file re-read error.");

	    (void)remove(path);

	    for (x=0; x<Ix; x++) {
		for (y=0; y<Ly; y++) W[y]=F[y][x];

		FFT(Ly,Ey,W,1);

		for (y=0; y<Ly; y++) F[y][x]=W[y];
	    }
	    for (y=0; y<Ly; y++) {
		for (x=0; x<Ix; x++) W[x]=F[y][x];
		for (; x<Lx; x++) {
		    W[x].r=F[y][Lx-x].r; W[x].i=(-F[y][Lx-x].i);
		}
		FFT(Lx,Ex,W,1);

		for (x=0; x<Lx; x++) P[y][x]=W[x].r/Lxyz;
	    }
	    if (z==0) {
		if (**P==0.0) Error("no object pixel.");

		(void)printf("%lf\n",
			     p=(double)**P/(double)Nx/(double)Ny/(double)Nz);

		q=(double)c/p;
	    }
	    r=q/(double)(Nz-z);

	    if (z!=0) {
		for (y=Y; y<=0; y++) {
		    s=r/(double)(Ny+y);
		    for (x=X; x<=0; x++) cell[y-Y][x-X]=PV(P[-y][-x],Nx+x);
		    for (; x<Dx; x++) cell[y-Y][x-X]=PV(P[-y][Lx-x],Nx-x);
		}
		for (; y<Dy; y++) {
		    s=r/(double)(Ny-y);
		    for (x=X; x<=0; x++) cell[y-Y][x-X]=PV(P[Ly-y][-x],Nx+x);
		    for (; x<Dx; x++) cell[y-Y][x-X]=PV(P[Ly-y][Lx-x],Nx-x);
		}
		(void)sprintf(path,tif,-z-Z);
		StoreImageFile(path,Mx,My,BPS,cell,NULL);
	    }
	    for (y=Y; y<0; y++) {
		s=r/(double)(Ny+y);
		for (x=X; x<0; x++) cell[y-Y][x-X]=PV(P[Ly+y][Lx+x],Nx+x);
		for (; x<Dx; x++) cell[y-Y][x-X]=PV(P[Ly+y][x],Nx-x);
	    }
	    for (; y<Dy; y++) {
		s=r/(double)(Ny-y);
		for (x=X; x<0; x++) cell[y-Y][x-X]=PV(P[y][Lx+x],Nx+x);
		for (; x<Dx; x++) cell[y-Y][x-X]=PV(P[y][x],Nx-x);
	    }
	    (void)sprintf(path,tif,z-Z);
	    StoreImageFile(path,Mx,My,BPS,cell,NULL);
	}
	return 0;
}
