
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"
#include "FFT.h"

extern void	Error(),RangeList();

static int	R_D(str,N)
char		*str;
int		N;
{
	int	D=(*str=='-')?(int)(atof(str+1)*(double)N+0.5):atoi(str);

	if (D<=0 || D>N) Error("bad correlation distance.");

	return D;
}

#define CELL	(size_t)((Cell)(~0))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,y,x,Dx,Dy,BPS,Mx,My,Lx,Ly,Ix,X,Y;
	Cell	**org,**new,c;
	char	binary[CELL+1];
	Complex	*Ex,*Ey,*W,**F;
	double	**P,Lxy,p,q,r,s;

	if (argc!=4 && argc!=5 && argc!=6 && argc!=7)
	    Error("usage : t_acc orgTIFF {rangeList} {Dx Dy} BPS newTIFF");

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&org,NULL);

	(void)printf("%d\t%d\n",Nx,Ny);

	if (argc==5 || argc==7) {
	    RangeList(argv[2],CELL,binary);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) org[y][x]=(Cell)binary[org[y][x]];
	}
	if (argc==4 || argc==5) {
	    Dx=Nx; Dy=Ny;
	}
	else {
	    Dx=R_D(argv[argc-4],Nx);
	    Dy=R_D(argv[argc-3],Ny);
	}
	(void)printf("%d\t%d\n",Dx,Dy);

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad BPS.");

	Mx=Dx*2-1; My=Dy*2-1;

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

	if ((Lx=SetUpFFT(Nx+Dx,&Ex))==0 ||
	    (Ly=SetUpFFT(Ny+Dy,&Ey))==0 ||
	    (W=ALLOC(Complex,(size_t)((Lx>Ly)?Lx:Ly)))==NULL ||
	    (F=ALLOC(Complex *,(size_t)Ny))==NULL ||
	    (*F=ALLOC(Complex,(size_t)Ny*(size_t)(Ix=Lx/2+1)))==NULL ||
	    (new=ALLOC(Cell *,(size_t)My))==NULL ||
	    (*new=ALLOC(Cell,(size_t)My*(size_t)Mx))==NULL)
	    Error("memory allocation error.");

	for (y=1; y<Ny; y++) F[y]=F[y-1]+Ix;

	for (y=1; y<My; y++) new[y]=new[y-1]+Mx;

	for (y=0; y<Ny; y++) {
	    for (x=0; x<Nx; x++) {
		W[x].r=(double)org[y][x]; W[x].i=0.0;
	    }
	    for (; x<Lx; x++) W[x].r=W[x].i=0.0;

	    FFT(Lx,Ex,W,-1);

	    for (x=0; x<Ix; x++) F[y][x]=W[x];
	}
	for (x=0; x<Ix; x++) {
	    for (y=0; y<Ny; y++) W[y]=F[y][x];
	    for (; y<Ly; y++) W[y].r=W[y].i=0.0;

	    FFT(Ly,Ey,W,-1);

	    for (y=0; y<Ly; y++) {
		W[y].r=W[y].r*W[y].r+W[y].i*W[y].i; W[y].i=0.0;
	    }
	    FFT(Ly,Ey,W,1);

	    for (y=0; y<Dy; y++) F[y][x]=W[y];
	}
	P=(double **)F; Lxy=(double)Lx*(double)Ly;

	for (y=0; y<Dy; y++) {
	    for (x=0; x<Ix; x++) W[x]=F[y][x];
	    for (; x<Lx; x++) {
		W[x].r=F[y][Lx-x].r; W[x].i=(-F[y][Lx-x].i);
	    }
	    FFT(Lx,Ex,W,1);

	    for (x=0; x<Lx; x++) P[y][x]=W[x].r/Lxy;
	}
	if (**P==0.0) Error("no object pixel.");

	(void)printf("%lf\n",p=(double)**P/(double)Nx/(double)Ny);

	X=1-Dx; Y=1-Dy; q=(double)(c=(Cell)((1<<BPS)-1))/p;

#define PV(Pxy,nx)	((s=Pxy*r/(double)(nx))<=(double)c)?(Cell)(s+0.5):c

	for (y=Y; y<0; y++) {
	    r=q/(double)(Ny+y);
	    for (x=X; x<=0; x++) new[y-Y][x-X]=PV(P[-y][-x],Nx+x);
	    for (; x<Dx; x++) new[y-Y][x-X]=PV(P[-y][Lx-x],Nx-x);
	}
	for (; y<Dy; y++) {
	    r=q/(double)(Ny-y);
	    for (x=X; x<0; x++) new[y-Y][x-X]=PV(P[y][Lx+x],Nx+x);
	    for (; x<Dx; x++) new[y-Y][x-X]=PV(P[y][x],Nx-x);
	}
	StoreImageFile(argv[argc-1],Mx,My,BPS,new,NULL);

	return 0;
}
